'' =================================================================================================
''
''   File....... jm_lanc.spin
''   Purpose.... LANC interface (for non-inverting circuit)
''   Author..... Jon "JonnyMac" McPhalen 
''               Copyright (c) 2014 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 07 APR 2014
''
'' =================================================================================================

{

  For suggested LANC-Propeller interface see:
  -- http://www.jonmcphalen.com/lanc_interface.jpg

  Excellent online LANC reference:
  -- http://www.boehmel.de/lanc.htm

  *** WARNING ***
  Avoid LANC PDF sold by Applied Logic Engineering
  -- material is OUTDATED
     * everything (and more) in the document is freely available online
  -- no need to waste $20
     * see above link for good information

}


con

  REPS = 4                                                      ' new command repititions
  

var

  long  cog                                                     ' cog running LANC
  long  cogparms[7]                                             ' parameters for LANC cog

  long  usercmd                                                 ' packed command
  long  framecount                                              ' lanc frames rx'd
  
  byte  buffer[8]                                               ' receive buffer 


pub start(sio)

'' Creates open-true LANC UART

  stop                                                          ' stop cog if running

  ' setup parameters for the cog
  
  cogparms[0] := 1 << sio                                       ' cog:siomask                 
  cogparms[1] := clkfreq / 9600                                 ' cog:bit1x0
  cogparms[2] := cogparms[1] * 15 / 10                          ' cog:bit1x5
  cogparms[3] := (clkfreq / 1000) << 1                          ' cog:idletix (2ms)         
  cogparms[4] := clkfreq >> 2                                   ' cog:lostix  (250ms)                    
  cogparms[5] := @framecount                                    ' cog:fcaddr
  cogparms[6] := @buffer                                        ' cog:bufaddr

  usercmd := @cogparms                                          ' point to parameters
  
  cog := cognew(@lancio, @usercmd) + 1                          ' launch the cog

  return cog     


pub stop

'' Stops LANC cog if running

  if (cog)
    cogstop(cog - 1)
    cog := 0

  bytefill(@buffer, $00, 8)                                     ' no data
  framecount := 0                                               ' no LANC (yet)


pub command(lanc1, lanc2) | uc

'' Start LANC command
'' -- sends REPS times
'' -- note: app should wait ~85ms before sending NEW command
''          * same command will update tx reps

  if (framecount == 0)                                          ' if no LANC
    return false                                                '  abort

  ' if busy, update count for same command
  ' -- disallow new command until not busy
  
  ifnot (busy)
    uc.byte[0] := REPS                                          ' build packed command
    uc.byte[1] := lanc1
    uc.byte[2] := lanc2
    uc.byte[3] := 0                                             ' not used at this time
     
    usercmd := uc                                               ' alert cog
    return true

  else  
    if (lanc1 <> usercmd.byte[1])
      return false                                              ' too early for new cmd  
    elseif (lanc2 <> usercmd.byte[2])
      return false
    else
      usercmd.byte[0] := REPS                                   ' reset reps count
      return true


pub wake_up

'' Wake sleeping camera

  if (framecount == 0)                                          ' if no LANC stream
    usercmd := $00_00_FF


pub available

'' Reports true when LANC stream detected

  return (framecount > 0)
  

pub frame

'' Returns frame counter
'' -- if 0, no frames recieved (or signal was lost)

  return framecount
  

pub busy

'' Reports true when command count > 0

  if (framecount == 0)
    return false
  else
    return (usercmd.byte[0] > 0)                                ' command count

  
pub address

'' Returns address of rx buffer

  return @buffer
  
  
dat

' Note: This code is desiged for use with a non-inverting, half-duplex level-shifter
'       circuit; LANC command codes (as found in Sony docs and online) are software
'       inverted for the interface.
'
'       http://www.jonmcphalen.com/lanc_interface.jpg
'
'       Remove lines marked with {*} if using inverting interface.

                        org     0

lancio                  rdlong  r1, par                         ' hub address of parameters -> t1
                        movd    :read, #siomask                 ' location of cog parameters -> :read(dest)
                        mov     r2, #7                          ' transfer 7 parameters
:read                   rdlong  0-0, r1                         ' copy parameter from hub to cog
                        add     r1, #4                          ' next hub element
                        add     :read, INC_DEST                 ' next cog element                         
                        djnz    r2, #:read                      ' done?

                        andn    outa, siomask                   ' 0 to sio bit 
                        andn    dira, siomask                   ' float to pull-up                 

                        mov     ctra, FREE_RUN                  ' set ctra for LOS timing
                        mov     frqa, #1 
                        mov     phsa, #0                        ' reset LOS timer

                        mov     lancframes, #0                  ' initial frame counter
                                                
                        
clear_cmd               mov     r1, #0
                        wrlong  r1, par

                        
lanc_loop               rdlong  cmd, par                        ' check user commad
                        mov     cmdcount, cmd
                        and     cmdcount, #$FF                  ' extract count
                        tjz     cmdcount, #rx_lanc8             ' no command

                        cmp     cmdcount, #$FF          wc, wz  ' wake-up?
        if_e            jmp     #wake
                        
                        mov     cmdb1, cmd
                        shr     cmdb1, #8
                        and     cmdb1, #$FF                     ' extract cmd byte 1

                        mov     cmdb2, cmd
                        shr     cmdb2, #16
                        and     cmdb2, #$FF                     ' extract cmd byte 2

                        
tx_lanc_cmd             call    #wait_idle

                        mov     txwork, cmdb1                   ' prep to tx cmdb1
                        call    #transmit
                        tjnz    losflag, #bad_packet            ' abort if lanc signal lost
                        mov     bufpntr, bufaddr
                        wrbyte  cmdb1, bufpntr                  ' cmdb1 --> buffer[0]
                        
                        mov     txwork, cmdb2
                        call    #transmit
                        tjnz    losflag, #bad_packet
                        add     bufpntr, #1  
                        wrbyte  cmdb2, bufpntr                  ' cmdb2 --> buffer[1]

                        mov     rxcount, #6                     ' rx 6 data bytes

:loop                   call    #receive                        ' rx a byte
                        add     bufpntr, #1                     ' update rx buffer pointer
                        wrbyte  rxwork, bufpntr                 ' write to hub
                        tjnz    losflag, #bad_packet            ' abort if signal lost
                        djnz    rxcount, #:loop

                        rdlong  r1, fcaddr                      ' read frame count
                        add     r1, #1                          ' increment
                        wrlong  r1, fcaddr                      ' write it back

                        rdbyte  r1, par                 wz      ' read cmd count
        if_nz           sub     r1, #1                          ' decrement
        if_nz           wrbyte  r1, par
                        jmp     #lanc_loop


rx_lanc8                call    #wait_idle
                        mov     bufpntr, bufaddr                ' point to buffer[0]
                        mov     rxcount, #8                     ' rx 8 bytes
                        
:loop                   call    #receive                        ' rx a byte
                        wrbyte  rxwork, bufpntr                 ' write to hub
                        add     bufpntr, #1                     ' increment buf pointer
                        tjnz    losflag, #bad_packet            ' abort if lanc signal lost
                        djnz    rxcount, #:loop

                        add     lancframes, #1                  ' increment frames
                        wrlong  lancframes, fcaddr              ' write to hub
                        jmp     #lanc_loop                             

                        
bad_packet              mov     lancframes, #0                  ' lancframes := 0                
                        wrlong  lancframes, fcaddr              ' write to frame count
                        mov     phsa, #0                        ' restart los timing                                                     
                        jmp     #lanc_loop 

                        
wake                    andn    outa, siomask                   ' sio low 
                        or      dira, siomask
                        mov     r1, lostix                      ' hold 1/4s
                        add     r1, cnt
                        waitcnt r1, #0
                        andn    dira, siomask                   ' float sio
                        mov     phsa, #0                        ' restart los timing                                                     
                        jmp     #clear_cmd


                        
' ==============================
' Wait for idle period
' ==============================
'
wait_idle               neg     r1, cnt                         ' r1 := -cnt
:loop                   test    siomask, ina            wc      ' sample rx
        if_nc           jmp     #wait_idle                      ' if bit, restart
                        mov     r2, r1                          ' copy start to r2
                        adds    r2, cnt                         ' r2 hold idle ticks
                        cmps    r2, idletix             wc, wz  ' compare to threshold
        if_b            jmp     #:loop                          ' if below, keep trying
wait_idle_ret           ret                                     ' idle period detected, exit



' ==============================
' Receive LANC byte
' -- byte returned in rxwork
' ==============================
'
receive                 mov     rxwork, #0                      ' clear work var
                        mov     rxbits, #8                      ' rx 8 bits
:loop                   mov     rxtimer, cnt                    ' sync bit timer
                        test    siomask, ina            wc      ' look for start bit
        if_nc           jmp     #rx_byte                        ' if c = 0, go!
                        
                        cmp     lostix, phsa            wc, wz  ' check LOS timer        
        if_a            jmp     #:loop                          ' if lostix > phsa, loop 

                        mov     losflag, #1                     ' note signal loss
                        jmp     #receive_ret                    ' and exit
                        
rx_byte                 add     rxtimer, bit1x5                 ' start timer
:loop                   waitcnt rxtimer, bit1x0                 ' hold for middle of bit
                        test    siomask, ina            wc      ' rx --> c
                        shr     rxwork, #1                      ' prep for new bit
                        muxc    rxwork, #%1000_0000             ' c --> rxwork.7
                        djnz    rxbits, #:loop                  ' update bit count
                        waitcnt rxtimer, #0                     ' let last bit finish
{*}                     xor     rxwork, #$FF                    ' invert bits
                        mov     phsa, #0                        ' reset los timer
                        mov     losflag, #0                     ' clear los flag     
receive_ret             ret  



' ==============================
' Transmit LANC byte
' -- byte in txwork
' -- requires external start bit
' ==============================
'
transmit                mov     txbits, #8                      ' tx 8 bits
{*}                     xor     txwork, #$FF                    ' invert bits 
:loop                   mov     txtimer, cnt                    ' sync bit timer
                        test    siomask, ina            wc      ' look for start bit
        if_nc           jmp     #txbyte                         ' if c = 0, go!

                        cmp     lostix, phsa            wc, wz  ' check LOS timer
        if_a            jmp     #:loop                          ' if lostix > phsa, loop

                        mov     losflag, #1                     ' note signal loss
                        jmp     #transmit_ret                   ' and exit

txbyte                  add     txtimer, bit1x0                 ' start timer
                        waitcnt txtimer, bit1x0                 ' skip start bit
                        
tx_bit                  test    txwork, #1              wc      ' get bit
                        muxnc   dira, siomask                   ' output the bit   
                        shr     txwork, #1                      ' prep for next bit
                        waitcnt txtimer, bit1x0                 ' let bit timer expire      
                        djnz    txbits, #tx_bit
                        andn    dira, siomask                   ' float sio
                        mov     phsa, #0                        ' reset los timer
                        mov     losflag, #0                     ' clear los flag     
transmit_ret            ret


' -------------------------------------------------------------------------------------------------

INC_DEST                long    1 << 9                          ' to increment D field 

FREE_RUN                long    %11111 << 26                    ' free-running counter 

siomask                 res     1                               ' mask for serial io pin
bit1x0                  res     1                               ' ticks in 1 bit
bit1x5                  res     1                               ' ticks in 1.5 bits
idletix                 res     1                               ' ticks for idle period detect
lostix                  res     1                               ' ticks for loss-of-signal detect
fcaddr                  res     1                               ' hub address of frame count
bufaddr                 res     1                               ' hub address of lanc buffer

lancframes              res     1                               ' frames rx'd since wake-up
losflag                 res     1                               ' loss-of-signal flag

cmd                     res     1                               ' packed command
cmdcount                res     1                               ' count for command bytes
cmdb1                   res     1                               ' command byte 1
cmdb2                   res     1                               ' command byte 2

bufpntr                 res     1                               ' working pointer to rx buffer

rxcount                 res     1                               ' bytes to rx
rxwork                  res     1                               ' workspace for rx
rxbits                  res     1                               ' bits to rx
rxtimer                 res     1                               ' rx bit timer

txwork                  res     1                               ' workspace for tx
txbits                  res     1                               ' bits to tx
txtimer                 res     1                               ' tx bit timer
                                                                
r1                      res     1                               ' temp registers
r2                      res     1 
                                 
                        fit     496
                        

dat

{{

  Terms of Use: MIT License 

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}                    